/* Chesapeake TTCP
 * Copyright 1997, Chesapeake Computer Consultants, Inc.
 */

import java.io.File;
import java.util.*;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import hypercast.HyperCastStatsException;
//import hypercast.I_StatisticsAccess;
//import hypercast.I_Stats;
//import hypercast.StatsElement;
import hypercast.StatsProcessor;

// Parameters for a TCP or UDP data transfer.
// Values here are used by the TransThread thread
//   for doing a transfer.
//
//class ParamsClass implements I_Stats, I_StatisticsAccess {
class ParamsClass
{
	// Constants
	public static final boolean RECEIVE = true;
	public static final boolean TRANSMIT = false;

	public static final int TCP = 0;
	public static final int UDP = 1;
	public static final int MCAST = 2;
	public static final int OVERLAY = 3;

	public static final boolean RUN = true;
	public static final boolean STOP = false;

	// Transfer Parameters
	private String Host;
	private Vector HostList;
	private String MAddr;
	private int Port;
	private boolean TransmitReceive;
	private int Protocol;
	private int BufSize;
	private int NumBuf;

	private boolean stop = true;
	private int RepeatTime = 0;

	private String LogFile = null;
	private String SeqFile = null;
	private String BWFile = null;
	private String DelayFile = null;
	private String EventFile = null;
	private String InterceptFile = null;
	private int BWPeriod = 0;

	// Transfer Results
	private long Start, End; // Epoc based times
	private int BufCount; // Number of buffers actually transferred
	private String Stats;

	private int Rate, EndRate, RateStep = 25;

	private String OLConfig = "none";

	private long TotalDelay = 0;
	
	//for experiment only
	public boolean FixKey=false;
	
	private String Session="Default";
	
	private String statsDir = "TTCP_DATA/";

	// Constructor
	ParamsClass() {
		// Set default values (the GUI will use accessor methods to
		// discover the values we set here.)
		Host = "127.0.0.1";
		HostList = new Vector();
		Port = 5001;
		TransmitReceive = RECEIVE;
		Protocol = TCP;
		BufSize = 512;
		NumBuf = 1024;
		MAddr = "224.228.19.79";

		LogFile = null;
		SeqFile = null;
		BWFile = null;
		DelayFile = null;
		BWPeriod = 0;

		Rate = 0;
		EndRate = 0;
		RateStep = 25;
		
		//this.initStatisticsStructure();
	}
	public void setTotalDelay(long dl) {
		TotalDelay = dl;
	}
	public long getDelay() {
		return TotalDelay;
	}
	public void setOLConfig(String f) {
		OLConfig = f;
	}
	public String getOLConfig() {
		return OLConfig;
	}

	public void setRepeatTime(int rt) {
		RepeatTime = rt;
	}
	public int getRepeatTime() {
		return RepeatTime;
	}

	public int getBWPeriod() {
		return BWPeriod;
	}
	public void setBWPeriod(int bp) {
		BWPeriod = bp;
	}
	public void setEventFile(String fname) {
		EventFile = fname;
	}
	public String getEventFile() {
		return EventFile;
	}
	public void setDelayFile(String fname) {
		DelayFile = fname;
	}
	public String getDelayFile() {
		return DelayFile;
	}
	public void setBWFile(String fname) {
		BWFile = fname;
	}
	public String getBWFile() {
		return BWFile;
	}
	public void setLogFile(String fname) {
		LogFile = fname;
	}
	public String getLogFile() {
		return LogFile;
	}
	public void setSeqFile(String fname) {
		SeqFile = fname;
	}
	public String getSeqFile() {
		return SeqFile;
	}
	public void setStop(boolean stp) {
		stop = stp;
	}
	public boolean getStop() {
		return stop;
	}
	public void setMCASTAddr(String ma) {
		MAddr = ma;
	}
	public String getMCASTAddr() {
		return MAddr;
	}
	// Accessor Functions
	public void addToHostList(String host) {
		HostList.addElement(host);
		System.out.println("send to " + host);
	}
	public String[] getHostList() {
		String[] results = new String[HostList.size()];
		for (int i = 0; i < results.length; i++)
			results[i] = (String) HostList.elementAt(i);

		return results;
	}
	public void setHost(String host) {
		Host = host;
	}
	public String getHost() {
		return Host;
	}
	public void setProtocol(int p) {
		Protocol = p;
	}
	public int getProtocol() {
		return Protocol;
	}
	public void setTR(boolean tr) {
		TransmitReceive = tr;
	}
	public boolean getTR() {
		return TransmitReceive;
	}
	public void setPort(int port) {
		Port = port;
	}
	public int getPort() {
		return Port;
	}
	public void setBufSize(int size) {
		BufSize = size;
	}
	public int getBufSize() {
		return BufSize;
	}
	public void setNumBuf(int count) {
		NumBuf = count;
	}
	public int getNumBuf() {
		return NumBuf;
	}

	public void setRate(int rate, int end_rate, int rate_step) {
		Rate = rate;
		EndRate = end_rate;
		RateStep = rate_step;
	}
	public int getRate() {
		return Rate;
	}
	public int getEndRate() {
		return EndRate;
	}
	public int getRateStep() {
		return RateStep;
	}

	public int BufSent = 0;

	public String getFormatedStatsHeader() {
		if (getTR())
			return "Session, Total packets, Total bytes, received, lost, Duration, Byte rate(k), Bit rate(k), Loss rate, Delivery Rate, Retransmission, Average Delay";
		else
			return "Session, Total packets, Total bytes, Duration, Byte rate(k), Bit rate(k), Total ACKs, ACK ratio, Average round-trip-time(ms)";
	}

	public String getFormatedStats(String session) {
		long Duration, Count;
		float KBps, fDuration, fCount;
		Duration = End - Start;

		float delay;
		if (BufCount <= 0)
			delay = 0;
		else
			delay = ((float) TotalDelay / (float) BufCount);

		if (getTR()) {

			Count = BufCount * BufSize;

			fCount = Count;
			fDuration = Duration;
			KBps = (fCount / fDuration);

			//System.out.println("NumBuf="+NumBuf);

			double lost_ratio = ((double) (NumBuf - BufCount)) / NumBuf;
			Stats =
				session
					+ ", "
					+ NumBuf
					+ ", "
					+ Count
					+ ", "
					+ BufCount
					+ ", "
					+ (NumBuf - BufCount)
					+ ", "
					+ Duration
					+ ", "
					+ KBps
					+ ", "
					+ 8 * KBps
					+ ", "
					+ lost_ratio
					+ ", "
					+ (1 - lost_ratio)
					+ ", "
					+ BufSent
					+ ", "
					+ delay;

			return Stats;
		} else {
			Count = NumBuf * BufSize;

			fCount = Count;
			fDuration = Duration;
			KBps = (fCount / fDuration);

			Stats =
				session
					+ ", "
					+ NumBuf
					+ ", "
					+ (NumBuf * BufSize)
					+ ", "
					+ Duration
					+ ", "
					+ KBps
					+ ", "
					+ 8 * KBps
					+ ", "
					+ BufCount
					+ ", "
					+ ((float) BufCount / (float) NumBuf)
					+ ", "
					+ delay;
			return Stats;
		}
	}
	public String getUnformatedStats() {
		long Duration, Count;
		float Kbps, fDuration, fCount;
		Duration = End - Start;

		float delay;
		if (BufCount <= 0)
			delay = 0;
		else
			delay = ((float) TotalDelay / (float) BufCount);

		if (TransmitReceive) {
			Count = BufCount * BufSize;
			fCount = Count;
			fDuration = Duration;
			Kbps = (fCount / fDuration);
			Stats = "Receive: ";
			double lost_ratio = 100.0 * (NumBuf - BufCount) / NumBuf;
			Stats =
				Stats
					+ (TransmitReceive ? "Receive: " : "Transmit: ")
					+ BufCount
					+ " packets"
					+ ", "+ 8*Kbps+" Kbps"
					+ ", lost "
					+ (NumBuf - BufCount)
					+ 
				//", lost ratio is " + (100*(NumBuf - BufCount)/NumBuf) + "%.";
	" packets, lost ratio is "
		+ lost_ratio
		+ "%, retransmission "
		+ BufSent
		+ "."
		+ "\naverage delay="
		+ delay;

		} else {
			Count = NumBuf * BufSize;
			fCount = Count;
			fDuration = Duration;
			Kbps = (fCount / fDuration);

			Stats =
				"Transmit: "
					+ NumBuf
					+ " packets in "
					+ Duration
					+ " milli-seconds = "
					+ Kbps
					+ " KB/sec ("
					+ 8 * Kbps
					+ " Kbps).\n"
					+ BufCount
					+ " ACKs recceived "
					+ "   Average round-trip-time="
					+ delay
					+ "ms";
		}
		return Stats;
	}
	public void setStart(long s) {
		Start = s;
	}
	public void setEnd(long e) {
		End = e;
	}
	public void setBufCount(int c) {
		BufCount = c;
	}

	StatsProcessor statsProcessor;

/*	
	private void initStatisticsStructure()
    {
        statsProcessor = new StatsProcessor();
		
        StatsElement se;

        se = new StatsElement ("Delay", null, this, true, true);
        se.setSchemaProperty ("xsd:Boolean", 0, 1, null, null);
        statsProcessor.addStatsElement (se);

        se = new StatsElement ("Stats", null, this, true, false);
        se.setSchemaProperty ("xsd:String", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("ConfigFile", null, this, true, true);
        se.setSchemaProperty ("xsd:String", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("Rate", null, this, true, true);
        se.setSchemaProperty ("xsd:Integer", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("NumBuf", null, this, true, true);
        se.setSchemaProperty ("xsd:Integer", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("BufSize", null, this, true, true);
        se.setSchemaProperty ("xsd:Integer", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("HostList", null, this, false, true);
        se.setSchemaProperty ("xsd:String", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
    }
*/
	
	/* (non-Javadoc)
	 * @see edu.virginia.cs.mng.hypercast.I_Stats#getStats(java.lang.String)
	 */
/*
	public String getStatsValue(String name) throws HyperCastStatsException {

		if(name.equals("Delay")){
			return ""+(this.getDelayFile()!=null);
		}else if (name.equals("Stats")) {
			return "\n" + this.getFormatedStatsHeader() + "\n" + this.getFormatedStats("NONE");
		} else if (name.equals("HostList")) {
			String hl = "";

			for (int i = 0; i < this.HostList.size(); i++) {
				hl += (String) HostList.elementAt(i);
				if (i < (this.HostList.size() - 1))
					hl += ";";
			}

			return hl;

		} else if (name.equals("ConfigFile")) {
			return getOLConfig();
		} else if (name.equals("Rate")) {
			return "" + getRate();
		} else if (name.equals("NumBuf")) {
			return "" + getNumBuf();
		} else if (name.equals("BufSize")){
			return "" + getBufSize(); 
		}
		
		else
			throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);
	}
*/

	String backupDelayFile=null;
	/* (non-Javadoc)
	 * @see edu.virginia.cs.mng.hypercast.I_Stats#setStats(java.lang.String, java.lang.String)
	 */
/*
	public String setStatsValue(String name, String value) throws HyperCastStatsException {
		
		if(name.equals("Delay")){
			boolean dl=false;
			if(value.equals("true"))dl=true;
			
			if(this.getDelayFile()!=null){
				if(!dl){
					backupDelayFile=this.getDelayFile();
					this.setDelayFile(null);
					System.out.println("Disable RTT measurement.");	
				}
			}else{
				if(dl){
					this.setDelayFile(backupDelayFile);
					System.out.println("Record RTT to: "+this.getDelayFile());
				}
			}
		}
		else if (name.equals("HostList")) {

			//System.out.println("set hostlist " + value);

			HostList = new Vector();
			String str = value.trim();

			if (str.equals("none"))
				str = "";

			while (str.length() > 0) {
				int index = str.indexOf(";");

				//System.out.println("str=" + str + " index=" + index);

				if (index < 0) {
					addToHostList(str);
					str = "";
				} else {

					addToHostList(str.substring(0, index));

					str = str.substring(index + 1);

				}
			}

		} else if (name.equals("ConfigFile")) {
			File cf = new File(value);
			if (!cf.exists())
				throw new HyperCastStatsException(name, HyperCastStatsException.InappropriateValue);

			setOLConfig(value);

		} else if (name.equals("Rate")) {
			int new_rate = Integer.parseInt(value);

			setRate(new_rate, new_rate, 0);
		} else if (name.equals("NumBuf")) {
			int new_nb = Integer.parseInt(value);
			setNumBuf(new_nb);
		} else if (name.equals("BufSize")){
			int new_bs = Integer.parseInt(value);
			setBufSize(new_bs); 
		} 
		else
			throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);
		
		return value;
	}
*/
	
	/**
	 * @return Returns the interceptFile.
	 */
	public String getInterceptFile() {
		return InterceptFile;
	}
	/**
	 * @param interceptFile The interceptFile to set.
	 */
	public void setInterceptFile(String interceptFile) {
		InterceptFile = interceptFile;
	}
	 /** @see I_Stats#getStats */
/*
    public Element getStats (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getStatsResult (doc, xpath);
    }
*/

    /** @see I_Stats#setStats */
/*
    public Element setStats (final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException
    {
        return statsProcessor.setStatsResult (doc, xpath, newValue);
    }
*/

    /** @see I_Stats#getReadSchema */
/*
    public Element getReadSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getReadSchemaResult (doc, xpath);
    }
*/

    /** @see I_Stats#getWriteSchema */
/*
    public Element getWriteSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getWriteSchemaResult (doc, xpath);
    }
*/

	/**
	 * @return Returns the session.
	 */
	public String getSession() {
		return Session;
	}
	/**
	 * @param session The session to set.
	 */
	public void setSession(String session) {
		Session = session;
	}
	/**
	 * @return Returns the statsDir.
	 */
	public String getStatsDir() {
		return statsDir;
	}
	/**
	 * @param statsDir The statsDir to set.
	 */
	public void setStatsDir(String statsDir) {
		this.statsDir = statsDir;
	}
}
