

import hypercast.ByteArrayUtility;

public class TTCPMessage {

	final static byte MSGTYPE_SYNC = 120;
	final static byte MSGTYPE_DATA = 121;
	final static byte MSGTYPE_ACK = 122;

	byte type;
	byte RIndex = 0;
	int size = 0;
	int numOfPacket = 0;
	int seqNum = 0;
	long timestamp = 0;
	boolean response = false;
	int ackSize = 0;
	//byte[] ipAddress = null;
	//int port = 0;

	long receivedTime = 0;
	
	String Session=null;

	public TTCPMessage(byte payload[]) {
		restoreMessage(payload);
	}

	public TTCPMessage(byte _type) {
		type = _type;
		switch (type) {
			case MSGTYPE_SYNC :
				size = 1 + 1;
				break;
			case MSGTYPE_DATA :
				size = 1 + 1 + 4 + 4 + 8 + 1;
				break;
			case MSGTYPE_ACK :
				size = 1 + 1 + 4 + 4 + 8 + 1 + 4 + 8;
				break;
			default :
				size = 1 + 1;
				break;
		}

		seqNum = 0;
		timestamp = 0;
		response = false;
	}
	public String getSession(){
		return Session;	
	}
	public void setSession(String ss){
		Session=ss;
		
	}
	
	public TTCPMessage(byte _type, int _size, int _seqNum) {
		type = _type;
		size = _size;
		seqNum = _seqNum;
		timestamp = 0;
	}
	public void setResponse(boolean _response) {
		response = _response;
	}
	public boolean getResponse() {
		return response;
	}
	public void setTimestamp(long _timestamp) {
		timestamp = _timestamp;
	}
	public long getTimestamp() {
		return timestamp;
	}
	public void setSeqNum(int _seqNum) {
		seqNum = _seqNum;
	}
	public int getSeqNum() {
		return seqNum;
	}
	public void setSize(int _size) {
		size = _size;
	}
	public int getSize() {
		return size;
	}
	public void setType(byte _type) {
		type = _type;
	}
	public byte getType() {
		return type;
	}
	public byte[] getPayload() {
		return toByteArray();
	}
	public void restoreMessage(byte payload[]) {
		int index = 0;

		type = 0;
		numOfPacket = 0;
		seqNum = -1;
		size = payload.length;
		timestamp = 0;
		response = false;
		ackSize = 0;
		
		
		if (payload.length < (index + 1))
			return;
		type = payload[index];
		index += 1;

		if (payload.length < (index + 1))
			return;
		RIndex = payload[index];
		index += 1;

		if (payload.length < (index + 4))
			return;
		numOfPacket = ByteArrayUtility.toInteger(payload, index);
		index += 4;

		if (payload.length < (index + 4))
			return;
		seqNum = ByteArrayUtility.toInteger(payload, index);
		index += 4;

		if (payload.length < (index + 8))
			return;
		timestamp = ByteArrayUtility.toLong(payload, index);
		index += 8;

		if (payload.length < (index + 1))
			return;
		byte res = payload[index];
		index += 1;
		if (res > 0)
			response = true;
		else
			response = false;

		if (payload.length < (index + 4))
			return;
		ackSize = ByteArrayUtility.toInteger(payload, index);
		index += 4;

		if (payload.length < (index + 8))
			return;
		receivedTime = ByteArrayUtility.toLong(payload, index);
		index += 8;
		
		if(payload.length < (index+4))
			return;
			
		int ll=ByteArrayUtility.toInteger(payload, index);
		index += 4;
		
			
		if(payload.length < (index+ll))
			return;
				
		Session=new String(payload, index, ll);
		index+=ll;
		
		

	}
	public byte[] toByteArray() {
		
		byte payload[] = new byte[size];
		int index = 0;

		if (size < (index + 1))
			return payload;
		payload[index] = type;
		index += 1;

		if (size < (index + 1))
			return payload;
		payload[index] = RIndex;
		index += 1;

		if (size < (index + 4))
			return payload;
		System.arraycopy(ByteArrayUtility.toByteArray(numOfPacket), 0, payload, index, 4);
		index += 4;

		if (size < (index + 4))
			return payload;
		System.arraycopy(ByteArrayUtility.toByteArray(seqNum), 0, payload, index, 4);
		index += 4;

		if (size < (index + 8))
			return payload;
		System.arraycopy(ByteArrayUtility.toByteArray(timestamp), 0, payload, index, 8);
		index += 8;

		if (size < (index + 1))
			return payload;
		byte res;
		if (response)
			res = 10;
		else
			res = 0;
		payload[index] = res;
		index += 1;

		if (size < (index + 4))
			return payload;
		System.arraycopy(ByteArrayUtility.toByteArray(ackSize), 0, payload, index, 4);
		index += 4;

		if (size < (index + 8))
			return payload;
		System.arraycopy(ByteArrayUtility.toByteArray(receivedTime), 0, payload, index, 8);
		index += 8;
		
		if(Session!=null){
			//System.out.println("copy Session");
			if (size >= (index + Session.length()+4)){
				System.arraycopy(ByteArrayUtility.toByteArray(Session.length()), 0, payload, index, 4);
				index+=4;
				
				System.arraycopy(Session.getBytes(), 0, payload, index, Session.length());
				index+=Session.length();
			}
		}else{
			//System.out.println("Session is null");
			
			if(size >=(index + 4)){
				System.arraycopy(ByteArrayUtility.toByteArray((int)0), 0, payload, index, 4);
				index+=4;
			}
			
		}

		return payload;
	}

}