/* Chesapeake TTCP
 * Copyright 1997, Chesapeake Computer Consultants, Inc.
 */

/*
  import java.net.ServerSocket;
  import java.net.Socket;
  import java.net.InetAddress;
  import java.io.DataInputStream;
  import java.io.DataOutputStream;
  import java.util.Date;
  import java.io.IOException;
  import java.io.EOFException;
  import java.net.SocketException;
 */
import java.net.*;
import java.io.*;
import java.util.*;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import hypercast.*;
//import hypercast.util.StatsUtil;
import hypercast.util.*;
import hypercast.events.NODE_NEIGHBORHOODCHANGED;
import hypercast.events.NOTIFICATION_EVENT;

import java.security.*;
import javax.crypto.spec.*;
import javax.crypto.KeyGenerator;


// Handle a TCP data transfer
//class TransThread implements I_ReceiveCallback, I_Stats, Runnable, I_InterceptionCallback, I_StatisticsAccess  {
class TransThread implements I_ReceiveCallback, Runnable, I_InterceptionCallback
{
	boolean monitor = false;

	// Local Variables
	private long start, end; // Epoch based start and end times
	private String Host;
	private int Port;
	private boolean TransmitReceive; // TRANSMIT or RECEIVE
	private int Protocol; // TCP or UDP or MCAST
	private int BufSize;
	private int NumBuf;
	private byte[] buf;
	private ServerSocket Server;
	private Socket Sock;
	private ParamsClass Params;
	private String maddr;

	private final int pkgs_onetime = 1; //pkgs sent by udp sender between sleeps
	private final int synchronization_try_times = 3;
	//tries for udp sender to synchronize with the receiver
	private DatagramSocket UDPSock;
	private DatagramSocket UDPSocks[];
	private MulticastSocket MSock;
	private int rate;
	private int end_rate;
	private int rate_step = 25;

	private final int MAX_BufSize = 300000;
	
	private StatsProcessor statsProcessor;
	private SocketStatsTable socketTable;
	
	
	private String State="NONE";
	
	EventHandler eh;

	TransThread(ParamsClass p) {
		/* XXX  Bogon alert!
		 * In an ideal world, ttcpControls should not be seen by
		 * TransThread.  Until I learn of another way to do
		 * interthread communication, we're going to see if
		 * Controls is non-null and use it to communicate with
		 * the GUI.  If Controls is null, then we're running
		 * without the GUI and the output goes directly to the
		 * console.  The function 'report()' does this work.
		 */
		Params = p;
		Host = p.getHost();
		Port = p.getPort();
		TransmitReceive = p.getTR();
		Protocol = p.getProtocol();
		BufSize = p.getBufSize();
		NumBuf = p.getNumBuf();
		maddr = p.getMCASTAddr();

		buf = new byte[MAX_BufSize];
		Server = null;
		Sock = null;

		UDPSock = null;
		MSock = null;
		rate = p.getRate();
		end_rate = p.getEndRate();
		
		eh=new EventHandler();
		//(new Thread(eh)).start();
		
		socketTable=new SocketStatsTable();
		
		//initStatisticsStructure();
		
	}
	

	int counter = 1;
	// Thread exection mainline
	String logheader = "";
	String prot_str;

	boolean transStarted = false;
	
	
	public void run() {
		int i;
		//System.out.println("monitor="+monitor+" TR="+Params.getTR());
		transStarted = true;

		switch (Protocol) {
			case 0 :
				prot_str = "TCP ";
				break;
			case 1 :
				prot_str = "UDP ";
				break;
			case 2 :
				prot_str = "Multicast ";
				break;
			case 3 :
				prot_str = "Overlay ";
				break;
			default :
				prot_str = "Unknown protocol";
				break;
		}
		logheader = prot_str + "  " + (new Date()) + "\n" + Params.getFormatedStatsHeader();
		if (!monitor || Params.getTR())
			saveStatsToFile(logheader);

		System.out.println("counter " + counter);

		NumBuf = Params.getNumBuf();

		while (true) {

			//openBWFile(counter+1);
			BufSize = Params.getBufSize();
			zeroTime = 0;
			lastTimeAckReceived = 0;
			
			State="Start";
			BufSize = Params.getBufSize();
			
			if (Protocol == ParamsClass.TCP) {
				TransTCP2();

			} else if (Protocol == ParamsClass.UDP) {
				TransUDP();

			} else if (Protocol == ParamsClass.MCAST) {
				/* multicast */

				TransMCAST();

			} else if (Protocol == ParamsClass.OVERLAY) {
				/* overlay */
				TransOverlay();

				if (toRestart) {
					toRestart = false;
					if (OLSock != null) {
						joined = false;
						OLSock.leaveOverlay();
						OLSock = null;
						//socketTable.clear();
					}
					if (!Params.getTR())
						initOLSocket(this);

					continue;
				}
			}

			//closeBWFile();
			int temp = Params.getNumBuf();
			Params.setNumBuf(NumBuf);

			System.out.println(Params.getUnformatedStats());

			saveStatsToFile(Params.getFormatedStats(session));
			Params.setNumBuf(temp);
			NumBuf = temp;

			//saveSequenceFile(counter);

			int repeatTime = Params.getRepeatTime();

			if (Params.getStop() || !(repeatTime == 0 || counter < repeatTime))
				break;
			if (TransmitReceive != ParamsClass.RECEIVE) {

				if (counter >= 100) {
					/* the maximum number of rounds for transmit */
					System.err.println("Maximum number of round reached.");
					break;
				}
				/* wait 10sec before next round */
				System.out.println("Wait 10sec before next round...");
				try {
					Thread.sleep(10000);
				} catch (Exception e) {
				}

			}
			counter++;
			System.out.println("\n\nStarting the " + (counter) + "th round");

		}
		counter++;
		transStarted = false;
	}

	
	
	//Vector bw_values;
	String data_dir = "TTCP_DATA/";
	
	String intercept_file;
	PrintWriter intercept_output = null;
	int max_TTL=255;
	
	private void openInterceptFile() {

		intercept_file = Params.getInterceptFile();
		if (intercept_file == null)
			return;

		intercept_file = data_dir + intercept_file;
		
		if(OLSock!=null)
			max_TTL= ((OL_Socket)this.OLSock).getHopLimit();
			
		try {
			intercept_output = new PrintWriter(new FileOutputStream(intercept_file));

			intercept_output.println(prot_str);
			intercept_output.println("Seq#\treceived\tsent time\tintercepted time\tSession\tSrc\tDest");

		} catch (Exception e) {
			System.err.println("Can't write to " + intercept_file + " : " + e.toString());

		}

	}
	
	private void saveIntercept(long now, I_OverlayMessage msg) {
		if (intercept_output == null){
			//System.out.println("Open event file the first time");
			openInterceptFile();
		}
		
		if (intercept_output == null){
			//System.out.println("Event file not open yet.");
			return;
		}

		try {
			//System.out.println("Event: "+event.get_event_info());
			TTCPMessage TtcpMsg=new TTCPMessage((byte)0);
			TtcpMsg.restoreMessage(msg.getPayload());
			if(TtcpMsg.getType()==TTCPMessage.MSGTYPE_DATA){
				String ss=TtcpMsg.getSession();
				if(ss!=null)
					session=ss;
				intercept_output.println(
						""+TtcpMsg.seqNum
						+"\t"+(max_TTL-msg.getHopLimit())
						+"\t"+TtcpMsg.getTimestamp()
						+"\t"+now
						+"\t"+session
						+"\t"+msg.getSourceAddress()+"\t"+((OL_Message) msg).getDestinationAddress()
				);
				intercept_output.flush();
			}

		} catch (Exception e) {
			System.err.println("Can't write to " + intercept_file + " : " + e.toString());

		}

	}
	private void closeInterceptFile() {
		try {
			if (intercept_output != null)
				intercept_output.close();
		} catch (Exception e) {
		}
		intercept_output=null;
	}

	
	PrintWriter event_output = null;
	String eventfile;
	private void openEventFile() {

		eventfile = Params.getEventFile();
		if (eventfile == null)
			return;

		eventfile = data_dir + eventfile;

		try {
			event_output = new PrintWriter(new FileOutputStream(eventfile));

			event_output.println(prot_str);
			event_output.println("time(ms)\tSession\tEvent");

		} catch (Exception e) {
			System.err.println("Can't write to " + eventfile + " : " + e.toString());

		}

	}
	
	private void saveEvent(long now, NOTIFICATION_EVENT event) {
		if (event_output == null){
			//System.out.println("Open event file the first time");
			openEventFile();
		}
		
		if (event_output == null){
			//System.out.println("Event file not open yet.");
			return;
		}

		try {
			//System.out.println("Event: "+event.get_event_info());
			event_output.println(""+now+"\t"+session+"\t"+event.getInfo());
			event_output.flush();

		} catch (Exception e) {
			System.err.println("Can't write to " + eventfile + " : " + e.toString());

		}

	}
	private void closeEventFile() {
		try {
			if (event_output != null)
				event_output.close();
		} catch (Exception e) {
		}
		event_output=null;
	}

	
	PrintWriter bw_output = null;
	String bw_file;
	private void openBWFile(int round) {

		bw_file = Params.getBWFile();
		if (bw_file == null)
			return;

		bw_file += "_" + round + ".txt";
		//bw_values=new Vector();
		bw_file = data_dir + session + "_" + bw_file;

		try {
			bw_output = new PrintWriter(new FileOutputStream(bw_file));

			bw_output.println(prot_str);
			bw_output.println("time(s)\tKBps\tKbps");

		} catch (Exception e) {
			System.err.println("Can't write to " + bw_file + " : " + e.toString());

		}

	}
	private void saveBW(int index, int packet_number, int bw_period) {

		/*
		float time=(index+1)*bw_period;
		float Bps=((float)packet_number*BufSize)/bw_period;
		float bps=Bps*8;System.out.println(""+index+":"+bps);
		*/

		if (bw_output == null)
			return;

		try {
			float time = (index + 1) * bw_period;
			float Bps = ((float) packet_number * BufSize) / bw_period;
			float bps = Bps * 8;

			bw_output.println(((float) time / 1000) + "\t" + Bps + "\t" + bps);
			//bw_values.addElement(""+((float)time/1000)+"\t"+Bps+"\t"+bps);

			//output.println("\n");

		} catch (Exception e) {
			System.err.println("Can't write to " + bw_file + " : " + e.toString());

		}

	}
	private void closeBWFile() {
		try {
			/*
			for(int i=0;i<bw_values.size();i++){
				bw_output.println((String)bw_values.elementAt(i));
			}
			bw_values=null;
			*/

			if (bw_output != null)
				bw_output.close();
		} catch (Exception e) {
		}

	}

	PrintWriter sq_output = null;
	String seqfile = null;

	private void openSQFile(int round) {

		totalDelay = 0;
		numOfAck = 0;
		numOfMsg = 0;

		seqfile = Params.getSeqFile();
		if (seqfile == null)
			return;

		sq_output = null;
		
		if(OLSock!=null)
			max_TTL= ((OL_Socket)this.OLSock).getHopLimit();
		
		seqfile += "_" + round + ".txt";

		seqfile = data_dir + session + "_" + seqfile;
		zeroTime = 0;

		try {
			sq_output = new PrintWriter(new FileOutputStream(seqfile));

			sq_output.println(prot_str);
			sq_output.println(
				"sequence#\treceived\tsent time\treceived time\tpseudo-delay\tafter reply");

		} catch (Exception e) {
			System.err.println("Can't write to " + seqfile + " : " + e.toString());

		}
	}
	long zeroTime = 0;
	private void saveSQ(int sq_num, long senttime, long time, OL_Message msg) {

		if (time == 0)
			time = System.currentTimeMillis();
		/*
		if (zeroTime == 0)
			zeroTime = time;
			*/
		
		totalDelay += (time - senttime);
		//numOfAck++;
		numOfMsg++;
		

		if (sq_output == null)
			return;
		
		int ttl=1;
		if(msg!=null){
			ttl=max_TTL-msg.getHopLimit();
		}

		try {

			sq_output.println(
				""
					+ sq_num
					+ "\t"
					+ ttl
					+ "\t"
					+ (senttime - zeroTime)
					+ "\t"
					+ (time - zeroTime)
					+ "\t"
					+ (time - senttime)
					+ "\t"
					+ (time_after_response - zeroTime));

		} catch (Exception e) {
			System.err.println("Can't write to " + seqfile + " : " + e.toString());

		}
	}
	private void closeSQFile() {
		try {

			Params.setTotalDelay(totalDelay);

			if (sq_output != null) {
				sq_output.close();

				//convert to density file
				//ttcp.sq2Ds(seqfile, seqfile + "_DS.txt", 50);
			}

		} catch (Exception e) {
		}

		sq_output = null;
	}

	/* save delay */
	PrintWriter delay_output = null;
	String delayfile = null;

	private void openDelayFile(int round) {
		delayfile = Params.getDelayFile();
		showAck=false;
		readAck=true;
		
		if (delayfile == null)
			return;

		delay_output = null;

		delayfile += "_" + round + ".txt";

		delayfile = data_dir + session + "_" + delayfile;

		zeroTime = 0;

		try {
			delay_output = new PrintWriter(new FileOutputStream(delayfile));

			delay_output.println(prot_str);
			delay_output.println(
				"receiver\tsequence#\tsize(byte)\tsent\treceived\tacked\tRTT\tdelay\tthrouput(kbps)");

		} catch (Exception e) {
			System.err.println("Can't write to " + delayfile + " : " + e.toString());

		}
	}

	private void saveDelay(
		String addr,
		int sq_num,
		int size,
		long sent,
		long acked,
		long received) {
		long RTT = (acked - sent);
		long delay = RTT / 2;
		totalDelay += (acked - sent);
		numOfAck++;
		/*
		if (zeroTime == 0)
			zeroTime = sent;
			*/
		
		if (delay_output == null)
			return;

		try {

			delay_output.println(
				addr
					+ "\t"
					+ sq_num
					+ "\t"
					+ size
					+ "\t"
					+ (sent - zeroTime)
					+ "\t"
					+ received
					+ "\t"
					+ (acked - zeroTime)
					+ "\t"
					+ RTT
					+ "\t"
					+ (delay)
					+ "\t"
					+ ((float) size * 8 / delay));

		} catch (Exception e) {
			System.err.println("Can't write to " + delayfile + " : " + e.toString());

		}

	}
	private void closeDelayFile() {
		try {
			if (delay_output != null)
				delay_output.close();

		} catch (Exception e) {
		}

		delay_output = null;
	}

	void saveStatsToFile(String stats) {
		String logfile = Params.getLogFile();
		if (logfile == null)
			return;
		PrintWriter output = null;

		logfile = data_dir + logfile;

		try {
			output = new PrintWriter(new FileOutputStream(logfile, true));

			output.println(stats);

			//output.println("\n");

		} catch (Exception e) {
			System.err.println("Can't write to " + logfile + " : " + e.toString());

		} finally {
			try {
				if (output != null)
					output.close();
			} catch (Exception e) {
			}
		}

	}

	int BW_PERIOD = 500; //check the bandwidth every half second
	long bw_starttime;
	long bw_endtime;
	int bw_index;

	int bw_packetcount;

	void initBWCheck() {

		BW_PERIOD = Params.getBWPeriod();
		if (BW_PERIOD <= 0) {
			BW_PERIOD = 500; //default value
			report("Bandwidth check interval set to default value: 500ms");
		}
		bw_starttime = System.currentTimeMillis();
		bw_endtime = bw_starttime;
		bw_packetcount = 0;
		bw_index = 0;

	}
	void checkBW() {
		
		if(this.bw_output==null)return;
		
		if ((end - bw_starttime) > ((bw_index + 1) * BW_PERIOD)) {
			//check the bandwidth
			//float bandwidth=(float)((bw_packetcount*BufSize)*8*1000)/BW_PERIOD;
			saveBW(bw_index, bw_packetcount, BW_PERIOD);

			bw_index++;
			bw_packetcount = 0;
		}
		bw_packetcount++;
		bw_endtime = System.currentTimeMillis();
	}

	private void responseTCPMessage(DataOutputStream out, TTCPMessage TtcpMsg, long now) {
		//System.out.println("send UDP response to "+UDPMsg.getAddress()+"/"+UDPMsg.getPort());

		TTCPMessage ResMsg = new TTCPMessage(TTCPMessage.MSGTYPE_ACK);
		ResMsg.setSeqNum(TtcpMsg.getSeqNum());
		ResMsg.setTimestamp(TtcpMsg.getTimestamp());
		ResMsg.ackSize = BufSize;
		
		/*
		if (zeroTime == 0)
			zeroTime = now;
			*/
		
		ResMsg.receivedTime = now - zeroTime;

		byte[] payload = ResMsg.getPayload();

		try {
			out.write(payload);
		} catch (IOException e) {
			e.printStackTrace();
		}

	}
	class TCPACKReceiver implements Runnable {
		boolean stopped = false;
		DataInputStream input;
		String host;
		int acksize;

		TCPACKReceiver(String h, DataInputStream in) {
			input = in;
			host = h;

			TTCPMessage TtcpMsg = new TTCPMessage(TTCPMessage.MSGTYPE_ACK);
			acksize = TtcpMsg.getSize();
		}

		public void stop() {
			stopped = true;
		}

		public void run() {
			TTCPMessage TtcpMsg = new TTCPMessage((byte) 0);

			while (!stopped) {
				try {
					input.readFully(buf, 0, acksize);

					TtcpMsg.restoreMessage(buf);

					//System.out.println("receive ACK from "+host+" type="+TtcpMsg.getType());

					if (TtcpMsg.getType() == TTCPMessage.MSGTYPE_ACK)
						ackReceived(host, TtcpMsg);

				} catch (InterruptedIOException e) {

				} catch (Exception e) {
					//e.printStackTrace();
					//System.out.println("ACK Receiver stopped.");
					break;
				}
			}
		}

	}
	//	Transfer using UDP
	private void TransTCP2() {
		int i = 0;
		long start_time = 0;
		long end_time = 0;
		long fix_sleep_time = 0;

		DataInputStream in = null;
		DataOutputStream out;
		DataOutputStream outs[] = null;
		DataInputStream ins[] = null;

		Socket Socks[] = null;
		
		State="Listening";

		if (TransmitReceive == ParamsClass.RECEIVE) {

			try {
				if (Server == null) { // Create receive server socket

					Server = new ServerSocket(Port);
				}
				report("TCP Receive: buflen= " + BufSize + "  nbuf= " + NumBuf + " port= " + Port);
				report("Wait to receive TCP packets:");
				Sock = Server.accept();
				
				report("Receive connection from:\n  " + Sock.toString() + ".");
				
				State="Receiving";
				
				in = new DataInputStream(Sock.getInputStream());
				out = new DataOutputStream(Sock.getOutputStream());

				TTCPMessage TtcpMsg = new TTCPMessage((byte) 0);

				report("Start to receive TCP packets:");

				openBWFile(counter);
				starttime();
				endtime();
				openSQFile(counter);

				initBWCheck();

				for (i = 1; i < NumBuf; i++) {

					in.readFully(buf, 0, BufSize);
					if (i == 1)
						starttime();

					endtime();

					TtcpMsg.restoreMessage(buf);

					if (TtcpMsg.getType() != TTCPMessage.MSGTYPE_DATA)
						continue;

					if (i == 1 && TtcpMsg.numOfPacket > 0) {
						NumBuf = TtcpMsg.numOfPacket;
						report("TCP Receive: nbuf adjusted to " + NumBuf);
					}

					int seq = TtcpMsg.getSeqNum();
					if (seq >= 0 && seq < NumBuf)
						saveSQ(seq, TtcpMsg.getTimestamp(), end, null);

					if (TtcpMsg.getResponse())
						responseTCPMessage(out, TtcpMsg, end);

					checkBW();

					if (i % 50 == 0) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
					
					if(seq>= (NumBuf-1)){
						/* the last message */
						break;
					}
					
				}
				System.out.println(".");
			} catch (SocketException e) {
				System.err.println("Receive: Socket exception: " + e);
			} catch (InterruptedIOException e) {
				//System.err.println("Receive: Timeout when receiving data: " + e);
				System.err.println("Receive: Receiving finishes but not all packets received.");
			} catch (IOException e) {
				System.err.println("Receive: IO exception: " + e);
			} finally {
				//endtime();
				
				State="End";
				
				closeBWFile();
				closeSQFile();
				savestats(i);

				try {
					if (Sock != null)
						Sock.close();
				} catch (Exception e) {
					report("Receive: socket close error: " + e);
				}
				Sock = null;
			}

		} else { // TRANSMIT
			
			
			if (rate == 0) {
				fix_sleep_time = 0;
			} else {
				fix_sleep_time = ((BufSize * 8 * pkgs_onetime)) * 1000 / (rate * NumBuf);
				System.out.println("fix_sleep_time=" + fix_sleep_time);
			}

			for (i = 0; i < BufSize; i++) {
				buf[i] = (byte) (i % 256);
			}
			String hostList[] = null;
			TCPACKReceiver ars[] = null;
			try {
				report("TCP Transmit: buflen= " + BufSize + "  nbuf= " + NumBuf + " port= " + Port);
				hostList = Params.getHostList();

				Socks = new Socket[hostList.length];
				outs = new DataOutputStream[hostList.length];
				ins = new DataInputStream[hostList.length];

				ars = new TCPACKReceiver[hostList.length];
				
				State="Connecting";

				for (int j = 0; j < hostList.length; j++) {
					try {

						Socks[j] = new Socket(hostList[j], Port);
						report("Transmit connection: " + Socks[j].toString() + ".");
						outs[j] = new DataOutputStream(Socks[j].getOutputStream());
						ins[j] = new DataInputStream(Socks[j].getInputStream());
						ars[j] = new TCPACKReceiver(hostList[j], ins[j]);
						Thread arthread = new Thread(ars[j]);
						//arthread.setPriority(Thread.MAX_PRIORITY);
						arthread.start();

					} catch (SocketException e) {
						report("Fail to establish connection: " + hostList[j] + ".");
						Socks[j] = null;
					}
				}

			} catch (SocketException e) {
				System.err.println("Transmit: Socket exception: " + e);

			} catch (UnknownHostException e) {
				System.err.println("Transmit: Unknown destination: " + e);
			} catch (IOException e) {
				System.err.println("Transmit: IO exception: " + e);
			}

			TTCPMessage TtcpMsg = new TTCPMessage(TTCPMessage.MSGTYPE_DATA, BufSize, 0);
			TtcpMsg.numOfPacket = NumBuf;
			if (Params.getDelayFile() != null)
				TtcpMsg.setResponse(true);

			openDelayFile(counter);

			try {
				report("Start to send TCP packets");
				starttime();

				rate = Params.getRate();
				end_rate = Params.getEndRate();
				rate_step = Params.getRateStep();
				int rate_interval = 0;

				if (end_rate < rate) {
					rate_step = -rate_step;
				}
				if (rate > 0 && rate_step != 0)
					rate_interval =
						NumBuf
							* BufSize
							* 8
							* 2
							/ (end_rate + rate)
							/ ((end_rate - rate) / rate_step + 1);

				int start_index = 0;

				//				sleep before sending
				try {
					report("Waiting 3sec before transmiting...");
					Thread.sleep(3000);
				} catch (Exception e) {

				}

				start_time = System.currentTimeMillis();
				starttime();
				totalDelay = 0;
				numOfAck = 0;
				
				openSQFile(counter);
				
				State="Transmitting";
				toStop=false;

				for (i = 0; i < NumBuf; i++) {
					//start_time = System.currentTimeMillis();
					if(toStop){
						throw new InterruptedException();
					}

					TtcpMsg.setSeqNum(i);
					TtcpMsg.setTimestamp(System.currentTimeMillis());
					byte[] payload = TtcpMsg.getPayload();

					for (int j = 0; j < hostList.length; j++) {
						try {

							if (Socks[j] != null)
								outs[j].write(payload);
						} catch (Exception e) {
							System.err.println(
								"Exception while sending to " + hostList[j] + ": " + e.toString());
						}

					}

					end_time = System.currentTimeMillis();
					saveSQ(i, end_time, end_time, null);

					if (rate > 0) {
						long next_time = (i - start_index + 1) * BufSize * 8 / rate;
						long sleep_time = next_time - (end_time - start_time);
						if ((rate > 0) && (i % pkgs_onetime == 0) && ((i - start_index) > 0)) {

							if (sleep_time > 0) {
								//System.out.println("sleep time=" + (fix_sleep_time- end_time + start_time));
								//Thread.sleep(fix_sleep_time - end_time + start_time - 10);
								Thread.sleep(sleep_time > 10 ? sleep_time - 10 : sleep_time);

							} else {
								//System.err.
								//println ("The specified sending rate is too fast!");
							}
							//System.out.println("sleep last time=" + (System.currentTimeMillis() - end_time));
						}

						if (rate_interval > 0 && (end_time - start_time) >= rate_interval) {
							rate += rate_step;
							report("Rate steps to " + rate + " Kbps");
							start_index = i;
							start_time = System.currentTimeMillis();
						}
					}

					if ((i % 50 == 0) && (i > 0)) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
				}
				System.out.println(".");
				System.out.println("finished.");

			} catch (InterruptedException e) {
				System.err.println("Receive: Sleep call failed.");
			} finally {
				State="Waiting ACK";
				
				endtime();
				savestats(numOfAck);

				if(lastTimeAckReceived<System.currentTimeMillis())
					lastTimeAckReceived=System.currentTimeMillis();
				
				showAck=true;
					
				/* sleep until no ack received anymore */
				while (showAck &&(System.currentTimeMillis() - lastTimeAckReceived) < sleepAfterSend) {
					try {
						Thread.sleep(2000);
					} catch (InterruptedException e1) {
						System.out.println("" + e1.toString());
					}
				}
				
				closeSQFile();
				
				report("\n");

				for (int j = 0; j < hostList.length; j++) {
					if (Socks[j] != null)
						ars[j].stop();
				}

				for (int j = 0; j < hostList.length; j++) {
					try {
						if (Socks[j] != null)
							Socks[j].close();
					} catch (Exception e) {
						report("Transmit: socket close error: " + e);
					}
					Socks[j] = null;
				}

				closeDelayFile();
				
				State="End";
			}
		}

	}
	// Transfer using TCP
	private void TransTCP() {
		int i = 0;
		DataInputStream in = null;
		DataOutputStream out = null;
		if (TransmitReceive == ParamsClass.RECEIVE) {
			try {
				if (Server == null) { // Create receive server socket

					Server = new ServerSocket(Port);
				}
				report("TCP Receive: buflen= " + BufSize + "  nbuf= " + NumBuf + " port= " + Port);
				Sock = Server.accept();

				openBWFile(counter);

				report("Receive connection from:\n  " + Sock.toString() + ".");
				in = new DataInputStream(Sock.getInputStream());

				initBWCheck();
				starttime();
				for (i = 0; i < NumBuf; i++) {
					in.readFully(buf, 0, BufSize);

					endtime();
					checkBW();

					if ((i % 50 == 0) && (i > 0)) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
				}

				closeBWFile();

				System.out.println(".");
			} catch (EOFException e) {
				report("Receive: EOF");
			} catch (Exception e) {
				report("Receive: socket or IO error: " + e);
			} finally {
				endtime();
				savestats(i);
				try {
					if (in != null)
						in.close();
					if (Sock != null)
						Sock.close();
					if (Server != null)
						Server.close();
				} catch (Exception e) {
					report("Receive: socket close error: " + e);
				}
				in = null;
				Sock = null;
				Server = null;
			}
		} else { // TRANSMIT
			for (i = 0; i < BufSize; i++) {
				buf[i] = (byte) (i % 256);
			}
			try {
				report("TCP Transmit: buflen= " + BufSize + "  nbuf= " + NumBuf + " port= " + Port);
				Sock = new Socket(Host, Port);
				report("Transmit connection:\n  " + Sock.toString() + ".");

				out = new DataOutputStream(Sock.getOutputStream());
				starttime();
				for (i = 0; i < NumBuf; i++) {
					out.write(buf, 0, BufSize);

					if ((i % 50 == 0) && (i > 0)) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
				}
				System.out.println(".");
			} catch (SocketException e) {
				report("Transmit: no receiver found listening");
				i = 0;
				starttime();
				start--;
			} catch (Exception e) {
				report("Transmitter socket or IO error: " + e);
			} finally {
				endtime();
				savestats(i);
				try {
					if (out != null)
						out.close();
					if (Sock != null)
						Sock.close();
				} catch (Exception e) {
					report("Transmitter socket close error: " + e);
				}
				out = null;
				Sock = null;
			}
		}
	}

	//InetAddress sender_addr = null;
	//int sender_port = 0;
	long time_after_response;
	private void responseUDPMessage(DatagramPacket UDPMsg, TTCPMessage TtcpMsg, long now) {
		//System.out.println("send UDP response to "+UDPMsg.getAddress()+"/"+UDPMsg.getPort());

		TTCPMessage ResMsg = new TTCPMessage(TTCPMessage.MSGTYPE_ACK);
		ResMsg.setSeqNum(TtcpMsg.getSeqNum());
		ResMsg.setTimestamp(TtcpMsg.getTimestamp());
		ResMsg.ackSize = UDPMsg.getLength();
		ResMsg.RIndex = TtcpMsg.RIndex;
		
		/*
		if (zeroTime == 0)
			zeroTime = now;
			*/
		
		ResMsg.receivedTime = now - zeroTime;

		byte[] payload = ResMsg.getPayload();

		DatagramPacket UDPpacket =
			new DatagramPacket(payload, payload.length, UDPMsg.getAddress(), UDPMsg.getPort());

		try {
			UDPSock.send(UDPpacket);

		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	//UDPACKSender udp_ack_sender;
	//DatagramSocket ack_udp_sock;

	class UDPACKSender extends Thread {
		Vector udpmsg_queue;
		Vector ttcpmsg_queue;
		Vector time_queue;
		public UDPACKSender() {
			udpmsg_queue = new Vector();
			ttcpmsg_queue = new Vector();
			time_queue = new Vector();
		}
		public void insertResponse(DatagramPacket UDPMsg, TTCPMessage TtcpMsg) {
			udpmsg_queue.insertElementAt(UDPMsg, 0);
			ttcpmsg_queue.insertElementAt(TtcpMsg, 0);
			time_queue.insertElementAt(new Long(System.currentTimeMillis()), 0);
			this.interrupt();
		}
		/* (non-Javadoc)
		 * @see java.lang.Runnable#run()
		 */
		public void run() {
			while (udpmsg_queue != null) {

				while (udpmsg_queue.size() > 0 && time_queue.size() > 0) {
					if (UDPSock == null) {
						udpmsg_queue = new Vector();
						ttcpmsg_queue = new Vector();
						continue;
					}

					long senttime = ((Long) time_queue.elementAt(0)).longValue();
					long now = System.currentTimeMillis();

					if ((now - senttime) < 200)
						continue;

					DatagramPacket dp = (DatagramPacket) udpmsg_queue.elementAt(0);
					TTCPMessage tm = (TTCPMessage) ttcpmsg_queue.elementAt(0);
					udpmsg_queue.removeElementAt(0);
					ttcpmsg_queue.removeElementAt(0);
					time_queue.removeElementAt(0);

					responseUDPMessage(dp, tm, now);
				}
				try {
					Thread.sleep(10000);
				} catch (InterruptedException e) {
					//e.printStackTrace();
				}

			}
		}

	}

	class UDPACKReceiver implements Runnable {
		boolean stopped = false;
		public void stop() {
			stopped = true;
		}
		public void run() {
			TTCPMessage TtcpMsg = new TTCPMessage((byte) 0);
			DatagramPacket UDPPack_R = new DatagramPacket(buf, buf.length);
			try {
				UDPSock.setSoTimeout(0);
			} catch (SocketException e1) {
				e1.printStackTrace();
			}

			while (!stopped) {

				try {
					//UDPSock.setSoTimeout(0);

					UDPSock.receive(UDPPack_R);
					TtcpMsg.restoreMessage(UDPPack_R.getData());

					//System.out.println("receive ACK from "+UDPPack_R.getAddress());

					if (TtcpMsg.getType() == TTCPMessage.MSGTYPE_ACK) {
						//ackReceived(UDPPack_R.getAddress().toString(), TtcpMsg);
						ackReceived(hostList[TtcpMsg.RIndex], TtcpMsg);
						//numOfAck++;
					}

				} catch (InterruptedIOException e) {

				} catch (Exception e) {
					//e.printStackTrace();
					System.out.println("ACK Receiver stopped.");
				}
			}
		}

	}

	// Transfer using UDP
	int Timeout = 20000;
	String hostList[];
	private void TransUDP() {
		int i = 0;
		long start_time = 0;
		long end_time = 0;
		long fix_sleep_time = 0;

		InetAddress dstaddr = null;
		InetAddress dstaddrs[] = null;
		hostList = null;
		
		

		if (TransmitReceive == ParamsClass.RECEIVE) {
			State="Wait";
			
			try {
				report("UDP Receive: buflen= " + BufSize + "  nbuf= " + NumBuf + " port= " + Port);
				UDPSock = new DatagramSocket(Port);

				//report("Sender Buffer="+UDPSock.getSendBufferSize());

				/*
				if(udp_ack_sender==null){
					udp_ack_sender=new UDPACKSender();
					//Thread uas=(new Thread(udp_ack_sender));
					//uas.setPriority(Thread.MIN_PRIORITY);
					//uas.start();
					udp_ack_sender.start();
				}
				*/

				//Synchronization time with receiver
				report("Wait to receive udp packets:");

				TTCPMessage TtcpMsg = new TTCPMessage((byte) 0);

				while (true) {
					DatagramPacket UDPPack_R = new DatagramPacket(buf, buf.length);

					UDPSock.receive(UDPPack_R);

					//System.out.println("The payload length is " + UDPPack_R.getLength());
					if (UDPPack_R.getLength() > 1) {
						starttime();
						endtime();
						openSQFile(counter);

						TtcpMsg.restoreMessage(UDPPack_R.getData());

						if (TtcpMsg.getType() == TTCPMessage.MSGTYPE_DATA) {
							int seq = TtcpMsg.getSeqNum();
							long received = System.currentTimeMillis();

							if (TtcpMsg.getResponse())
								responseUDPMessage(UDPPack_R, TtcpMsg, received);
							//udp_ack_sender.insertResponse(UDPPack_R, TtcpMsg);

							time_after_response = System.currentTimeMillis();

							if (seq >= 0 && seq < NumBuf)
								saveSQ(seq, TtcpMsg.getTimestamp(), received, null);

							//if(TtcpMsg.getResponse())responseOLMessage(RecvdMessage, TtcpMsg);
							if (TtcpMsg.numOfPacket > 0) {
								NumBuf = TtcpMsg.numOfPacket;
								report("Unicast Receive: nbuf adjusted to " + NumBuf);
							}

						}

						break;
					} else {
						starttime();
					}

					byte[] buf_R = new byte[1];
					InetAddress clientAddr = UDPPack_R.getAddress();
					int clientPort = UDPPack_R.getPort();
					UDPPack_R = new DatagramPacket(buf_R, buf_R.length, clientAddr, clientPort);

					UDPSock.send(UDPPack_R);
				}

				//starttime();
				stopReceiving = false;
				if (autoStop)
					UDPSock.setSoTimeout(Timeout);
				else
					UDPSock.setSoTimeout(2000);

				report("Start to receive udp packets:");

				openBWFile(counter);
				//openSQFile(counter);

				//int seq=getSequenceNumber(buf, 0);
				//if(seq>=0 && seq<NumBuf)saveSQ(seq, 0);

				initBWCheck();
				
				State="Receiving";

				for (i = 1; i < NumBuf; i++) {
					DatagramPacket UDPPack = new DatagramPacket(buf, buf.length);

					try {
						UDPSock.receive(UDPPack);

					} catch (InterruptedIOException e) {
						if (autoStop || stopReceiving)
							throw new InterruptedIOException();
						else
							continue;
					}

					endtime();
					TtcpMsg.restoreMessage(UDPPack.getData());
					if (TtcpMsg.getType() != TTCPMessage.MSGTYPE_DATA)
						continue;
					int seq = TtcpMsg.getSeqNum();

					if (TtcpMsg.getResponse())
						responseUDPMessage(UDPPack, TtcpMsg, end);
					//udp_ack_sender.insertResponse(UDPPack, TtcpMsg);

					time_after_response = System.currentTimeMillis();

					if (seq >= 0 && seq < NumBuf)
						saveSQ(seq, TtcpMsg.getTimestamp(), end, null);

					//if(TtcpMsg.getResponse())responseOLMessage(RecvdMessage, TtcpMsg);

					//seq=getSequenceNumber(buf, 0);
					//if(seq>=0 && seq<NumBuf)saveSQ(seq, end);

					/*
					  if (i == 0) {
					  report("Start to receive udp packets:");
					  //starttime();
					  //set timeout limit to 5 seconds for following receive() calls
					  UDPSock.setSoTimeout(5000);
					  }
					 */

					checkBW();

					//if ((i%50 == 0) && (i > 0)) {
					if (i % 50 == 0) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
					
					if(seq>= (NumBuf-1)){
						/* the last message */
						break;
					}
				}
				System.out.println(".");
			} catch (SocketException e) {
				System.err.println("Receive: Socket exception: " + e);
			} catch (InterruptedIOException e) {
				//System.err.println("Receive: Timeout when receiving data: " + e);
				System.err.println("Receive: Receiving finishes but not all packets received.");
			} catch (IOException e) {
				System.err.println("Receive: IO exception: " + e);
			} finally {
				//endtime();
				closeBWFile();
				closeSQFile();
				savestats(i);
				try {
					if (UDPSock != null)
						UDPSock.close();
				} catch (Exception e) {
					report("Receive: socket close error: " + e);
				}
				UDPSock = null;
				
				State="End";
			}

		} else { // TRANSMIT

			if (rate == 0) {
				fix_sleep_time = 0;
			} else {
				fix_sleep_time = ((BufSize * 8 * pkgs_onetime)) * 1000 / (rate * NumBuf);
				System.out.println("fix_sleep_time=" + fix_sleep_time);
			}

			for (i = 0; i < BufSize; i++) {
				buf[i] = (byte) (i % 256);
			}
			try {
				report("UDP Transmit: buflen= " + BufSize + "  nbuf= " + NumBuf + " port= " + Port);
				hostList = Params.getHostList();

				UDPSock = new DatagramSocket();

				dstaddrs = new InetAddress[hostList.length];

				for (int j = 0; j < dstaddrs.length; j++) {
					dstaddrs[j] = InetAddress.getByName(hostList[j]);
				}

				//dstaddr = InetAddress.getByName (Host);

				UDPSock.setSoTimeout(2000);
			} catch (SocketException e) {
				System.err.println("Transmit: Socket exception: " + e);
			} catch (UnknownHostException e) {
				System.err.println("Transmit: Unknown destination: " + e);
			}

			//Synchronization time with receiver
			int try_times = 0;
			int successful = 0;
			/*
			while (true)
			    {
			    	
			        try
			            {
			                
			                
			                byte[]buf_S = new byte[1];
			                DatagramPacket UDPpacket_T =
			                    new DatagramPacket (buf_S, buf_S.length, dstaddr, Port);
			                DatagramPacket UDPPack_T = new DatagramPacket (buf, buf.length);
			                UDPSock.send (UDPpacket_T);
			
			
			                
			                UDPSock.receive (UDPPack_T);
			
			                InetAddress ReceiverAddr = UDPPack_T.getAddress ();
			
			                if (dstaddr.equals (ReceiverAddr))
			                {
			                  successful = 1;
			                  //report("Transmit: Synchronization is successful!");
			                }
			                
			
			            }
			        catch (IOException e)
			            {
			                //System.err.println("Transmit: IO exception: " + e);
			                report ("Transmit: Synchronization try " + try_times +
			                        " failed!");
			                
			            }
			        finally
			            {
			                if (successful > 0)
			                    {
			                        break;
			                    }
			                else if (try_times < synchronization_try_times)
			                    {
			                        try_times++;
			                        continue;
			                    }
			                else
			                    {
			                        report ("Transmit: no receiver found listening");
			                    }
			
			                try
			                    {
			                        if (UDPSock != null)
			                            UDPSock.close ();
			                    }
			                catch (Exception e)
			                    {
			                        report ("Transmit: socket close error: " + e);
			                    }
			                //UDPSock = null;
			                return;
			            }
			    }
			*/

			TTCPMessage TtcpMsg = new TTCPMessage(TTCPMessage.MSGTYPE_DATA, BufSize, 0);
			TtcpMsg.numOfPacket = NumBuf;
			if (Params.getDelayFile() != null)
				TtcpMsg.setResponse(true);

			openDelayFile(counter);
			UDPACKReceiver ackreceiver = new UDPACKReceiver();
			Thread arthread = new Thread(ackreceiver);
			//arthread.setPriority(Thread.MAX_PRIORITY);
			arthread.start();

			try {
				report("Start to send udp packets to " + Host);
				starttime();

				rate = Params.getRate();
				end_rate = Params.getEndRate();
				rate_step = Params.getRateStep();
				int rate_interval = 0;

				if (end_rate < rate) {
					rate_step = -rate_step;
				}
				if (rate > 0 && rate_step != 0)
					rate_interval =
						NumBuf
							* BufSize
							* 8
							* 2
							/ (end_rate + rate)
							/ ((end_rate - rate) / rate_step + 1);

				int start_index = 0;

				//				sleep before sending
				try {
					report("Waiting 3sec before transmiting...");
					Thread.sleep(3000);
				} catch (Exception e) {

				}
				//System.out.println("Rate="+rate);

				start_time = System.currentTimeMillis();
				starttime();
				totalDelay = 0;
				numOfAck = 0;
				
				openSQFile(counter);
				
				State="Transmitting";
				toStop=false;
				
				for (i = 0; i < NumBuf; i++) {
					//start_time = System.currentTimeMillis();
					
					if(toStop){
						throw new InterruptedException();
					}
					
					TtcpMsg.setSeqNum(i);
					TtcpMsg.setTimestamp(System.currentTimeMillis());

					//DatagramPacket UDPpacket =
					//new DatagramPacket (buf, buf.length, dstaddr, Port);

					for (int j = 0; j < dstaddrs.length; j++) {
						TtcpMsg.RIndex = (byte) j;

						byte[] payload = TtcpMsg.getPayload();

						DatagramPacket UDPpacket =
							new DatagramPacket(payload, payload.length, dstaddrs[j], Port);

						/* sequence number */
						//setSequenceNumber(i, buf, 0);

						UDPSock.send(UDPpacket);
					}

					end_time = System.currentTimeMillis();
					this.saveSQ(i, end_time, end_time, null);

					if (rate > 0) {
						long next_time = (i - start_index + 1) * BufSize * 8 / rate;
						long sleep_time = next_time - (end_time - start_time);
						//System.out.println("sleep time="+sleep_time);

						if ((rate > 0) && (i % pkgs_onetime == 0) && ((i - start_index) > 0)) {

							if (sleep_time > 0) {
								//System.out.println("sleep time=" + (fix_sleep_time- end_time + start_time));
								//Thread.sleep(fix_sleep_time - end_time + start_time - 10);
								Thread.sleep(sleep_time > 10 ? sleep_time - 10 : sleep_time);

							} else {
								//System.err.
								//println ("The specified sending rate is too fast!");
							}
							//System.out.println("sleep last time=" + (System.currentTimeMillis() - end_time));
						}

						if (rate_interval > 0 && (end_time - start_time) >= rate_interval) {
							rate += rate_step;
							report("Rate steps to " + rate + " Kbps");
							start_index = i;
							start_time = System.currentTimeMillis();
						}
					}

					if ((i % 50 == 0) && (i > 0)) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
				}
				System.out.println(".");
			} catch (IOException e) {
				System.err.println("Transmit: IO exception: " + e);
			} catch (InterruptedException e) {
				System.err.println("Receive: Sleep call failed.");
			} finally {
				
				State="Waiting ACK";
				
				endtime();

				if(lastTimeAckReceived<System.currentTimeMillis())
					lastTimeAckReceived=System.currentTimeMillis();
				
				showAck=true;
				
				/* sleep until no ack received anymore */
				while (showAck && (System.currentTimeMillis() - lastTimeAckReceived) < sleepAfterSend) {
					try {
						Thread.sleep(2000);
					} catch (InterruptedException e1) {
						System.out.println("" + e1.toString());
					}
				}
				
				closeSQFile();

				savestats(numOfAck);

				ackreceiver.stop();
				
				report("\n");
				
				try {
					if (UDPSock != null)
						UDPSock.close();
				} catch (Exception e) {
					report("\nTransmit: socket close error: " + e);
				}
				UDPSock = null;

				closeDelayFile();
				
				State="End";
				
			}
		}

	}

	final static int SEQ_PER_BYTE = 128;

	void setSequenceNumber(int seq, byte buf[], int index) {
		if ((buf.length - index) < 4)
			return;
		buf[index + 0] = (byte) (seq % SEQ_PER_BYTE);
		seq /= SEQ_PER_BYTE;
		buf[index + 1] = (byte) (seq % SEQ_PER_BYTE);
		seq /= SEQ_PER_BYTE;
		buf[index + 2] = (byte) (seq % SEQ_PER_BYTE);
		seq /= SEQ_PER_BYTE;
		buf[index + 3] = (byte) (seq);

	}
	int getSequenceNumber(byte buf[], int index) {

		if ((buf.length - index) < 4)
			return -1;

		int seq;
		seq = buf[index + 3];
		seq = seq * SEQ_PER_BYTE + buf[index + 2];
		seq = seq * SEQ_PER_BYTE + buf[index + 1];
		seq = seq * SEQ_PER_BYTE + buf[index + 0];

		return seq;
	}

	// Transfer using Multicast
	private void TransMCAST() {
		int i = 0;
		long start_time = 0;
		long end_time = 0;
		long fix_sleep_time = 0;

		InetAddress dstaddr = null;
		
		
		if (TransmitReceive == ParamsClass.RECEIVE) {
			
			State="Wait";

			try {
				report(
					"Multicast Receive: buflen= "
						+ BufSize
						+ "  nbuf= "
						+ NumBuf
						+ " port= "
						+ Port);
				MSock = new MulticastSocket(Port);
				MSock.joinGroup(InetAddress.getByName(maddr));

				//Synchronization time with receiver
				report("Wait to receive multicast packets:");

				TTCPMessage TtcpMsg = new TTCPMessage((byte) 0);

				while (true) {
					DatagramPacket UDPPack_R = new DatagramPacket(buf, buf.length);

					MSock.receive(UDPPack_R);

					//System.out.println("The payload length is " + UDPPack_R.getLength());
					if (UDPPack_R.getLength() > 1) {
						starttime();
						endtime();
						openSQFile(counter);

						TtcpMsg.restoreMessage(UDPPack_R.getData());
						if (TtcpMsg.getType() == TTCPMessage.MSGTYPE_DATA) {
							int seq = TtcpMsg.getSeqNum();
							if (seq >= 0 && seq < NumBuf)
								saveSQ(seq, TtcpMsg.getTimestamp(), 0, null);
							//if(TtcpMsg.getResponse())responseOLMessage(RecvdMessage, TtcpMsg);
							if (TtcpMsg.numOfPacket > 0) {
								NumBuf = TtcpMsg.numOfPacket;
								report("Multicast Receive: nbuf adjusted to " + NumBuf);
							}

						}

						break;
					} else {
						starttime();
					}
					/*
					byte[]buf_R = new byte[1];
					InetAddress clientAddr = UDPPack_R.getAddress ();
					int clientPort = UDPPack_R.getPort ();
					UDPPack_R = new DatagramPacket (buf_R,
					        buf_R.length, clientAddr,
					        clientPort);
					
					MSock.send (UDPPack_R);
					*/
				}

				//starttime();

				//int seq=getSequenceNumber(buf, 0);
				//if(seq>=0 && seq<NumBuf)saveSQ(seq, 0);
				stopReceiving = false;
				if (autoStop)
					MSock.setSoTimeout(Timeout);
				else
					MSock.setSoTimeout(2000);

				openBWFile(counter);
				initBWCheck();

				report("Start to receive multicast packets:");
				
				State="Receiving";
				
				for (i = 1; i < NumBuf; i++) {
					DatagramPacket UDPPack = new DatagramPacket(buf, buf.length);

					try {
						MSock.receive(UDPPack);

					} catch (InterruptedIOException e) {
						if (autoStop || stopReceiving)
							throw new InterruptedIOException();
						else
							continue;
					}

					endtime();

					TtcpMsg.restoreMessage(UDPPack.getData());
					if (TtcpMsg.getType() != TTCPMessage.MSGTYPE_DATA)
						continue;
					int seq = TtcpMsg.getSeqNum();
					if (seq >= 0 && seq < NumBuf)
						saveSQ(seq, TtcpMsg.getTimestamp(), end, null);
					//if(TtcpMsg.getResponse())responseOLMessage(RecvdMessage, TtcpMsg);

					//seq=getSequenceNumber(buf, 0);
					//if(seq>=0 && seq<NumBuf)saveSQ(seq, end);
					/*
					  if (i == 0) {
					  report("Start to receive udp packets:");
					  //starttime();
					  //set timeout limit to 5 seconds for following receive() calls
					  UDPSock.setSoTimeout(5000);
					  }
					 */

					checkBW();

					//if ((i%50 == 0) && (i > 0)) {
					if (i % 50 == 0) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
					
					if(seq>= (NumBuf-1)){
						/* the last message */
						break;
					}
					
				}
				System.out.println(".");
			} catch (SocketException e) {
				System.err.println("Receive: Socket exception: " + e);
			} catch (InterruptedIOException e) {
				//System.err.println("Receive: Timeout when receiving data: " + e);
				System.err.println("Receive: Receiving finishes but not all packets received.");
			} catch (IOException e) {
				System.err.println("Receive: IO exception: " + e);
			} finally {
				//endtime();
				closeBWFile();
				closeSQFile();

				savestats(i);
				try {
					if (MSock != null)
						MSock.close();
				} catch (Exception e) {
					report("Receive: socket close error: " + e);
				}
				MSock = null;
				
				State="End";
			}

		} else { // TRANSMIT Multicast

			if (rate == 0) {
				fix_sleep_time = 0;
			} else {
				fix_sleep_time = (new Float((BufSize * 8 * pkgs_onetime) / rate)).longValue();
				//System.out.println("fix_sleep_time=" + fix_sleep_time);
			}

			for (i = 0; i < BufSize; i++) {
				buf[i] = (byte) (i % 256);
			}
			try {
				report("UDP Transmit: buflen= " + BufSize + "  nbuf= " + NumBuf + " port= " + Port);
				MSock = new MulticastSocket();
				MSock.joinGroup(InetAddress.getByName(Host));

				dstaddr = InetAddress.getByName(Host);

				MSock.setSoTimeout(2000);
			} catch (SocketException e) {
				System.err.println("Transmit: Socket exception: " + e);
			} catch (UnknownHostException e) {
				System.err.println("Transmit: Unknown destination: " + e);
			} catch (IOException e) {
				System.err.println("Transmit: IO error: " + e);
				System.exit(-1);
			}

			//Synchronization time with receiver
			int try_times = 0;
			int successful = 0;
			while (true) {
				try {
					byte[] buf_S = new byte[1];
					DatagramPacket UDPpacket_T =
						new DatagramPacket(buf_S, buf_S.length, dstaddr, Port);
					DatagramPacket UDPPack_T = new DatagramPacket(buf, buf.length);

					/* send three time to make sure the receiver receives */
					MSock.send(UDPpacket_T);
					MSock.send(UDPpacket_T);
					MSock.send(UDPpacket_T);

					/* don't synchronization 
					 * because it's multicast
					 */
					successful = 1;
					/*
					  UDPSock.receive (UDPPack_T);
					
					  InetAddress ReceiverAddr = UDPPack_T.getAddress ();
					
					  if (dstaddr.equals (ReceiverAddr))
					  {
					  successful = 1;
					  //report("Transmit: Synchronization is successful!");
					  }
					 */

				} catch (IOException e) {
					//System.err.println("Transmit: IO exception: " + e);
					report("Transmit: Synchronization try " + try_times + " failed!");
					/*
					  } catch(InterruptedIOException e) { 
					  //System.err.println("Transmit: Timeout when receiving data: " + e);
					  report("Transmit: no receiver found listening");
					 */
				} finally {
					if (successful > 0) {
						break;
					} else if (try_times < synchronization_try_times) {
						try_times++;
						continue;
					} else {
						report("Transmit: no receiver found listening");
					}

					try {
						if (MSock != null)
							MSock.close();
					} catch (Exception e) {
						report("Transmit: socket close error: " + e);
					}
					//UDPSock = null;

				}
				return;
			}

			TTCPMessage TtcpMsg = new TTCPMessage(TTCPMessage.MSGTYPE_DATA, BufSize, 0);
			TtcpMsg.numOfPacket = NumBuf;
			if (Params.getDelayFile() != null)
				TtcpMsg.setResponse(true);

			try {
				report("Start to send multicast packets to " + Host);
				starttime();

				rate = Params.getRate();
				end_rate = Params.getEndRate();
				rate_step = Params.getRateStep();
				int rate_interval = 0;

				if (end_rate < rate) {
					rate_step = -rate_step;
				}
				if (rate > 0 && rate_step != 0)
					rate_interval =
						NumBuf
							* BufSize
							* 8
							* 2
							/ (end_rate + rate)
							/ ((end_rate - rate) / rate_step + 1);

				int start_index = 0;
				//				sleep before sending
				try {
					report("Waiting 3sec before transmiting...");
					Thread.sleep(3000);
				} catch (Exception e) {

				}

				start_time = System.currentTimeMillis();
				openSQFile(counter);
				
				State="Transmitting";
				toStop=false;

				for (i = 0; i < NumBuf; i++) {
					//start_time = System.currentTimeMillis();
					if(toStop){
						throw new InterruptedException();
					}
					
					TtcpMsg.setSeqNum(i);
					TtcpMsg.setTimestamp(System.currentTimeMillis());
					byte[] payload = TtcpMsg.getPayload();

					//DatagramPacket UDPpacket =
					//new DatagramPacket (buf, buf.length, dstaddr, Port);

					DatagramPacket UDPpacket =
						new DatagramPacket(payload, payload.length, dstaddr, Port);

					/* sequence number */
					//setSequenceNumber(i, buf, 0);

					MSock.send(UDPpacket);

					end_time = System.currentTimeMillis();
					
					saveSQ(i, end_time, end_time, null);
					
					if (rate > 0) {
						long next_time = (i - start_index + 1) * BufSize * 8 / rate;
						long sleep_time = next_time - (end_time - start_time);
						if ((rate > 0) && (i % pkgs_onetime == 0) && ((i - start_index) > 0)) {

							if (sleep_time > 0) {
								//System.out.println("sleep time=" + (fix_sleep_time- end_time + start_time));
								//Thread.sleep(fix_sleep_time - end_time + start_time - 10);
								Thread.sleep(sleep_time > 10 ? sleep_time - 10 : sleep_time);
							} else {
								//System.out.println ("The specified sending rate is too fast!");
							}
							//System.out.println("sleep last time=" + (System.currentTimeMillis() - end_time));

							if (rate_interval > 0 && (end_time - start_time) >= rate_interval) {
								rate += rate_step;
								report("Rate steps to " + rate + " Kbps");
								start_index = i;
								start_time = System.currentTimeMillis();
							}
						}
					}

					if ((i % 50 == 0) && (i > 0)) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
				}
				System.out.println(".");
			} catch (IOException e) {
				System.err.println("Transmit: IO exception: " + e);
			} catch (InterruptedException e) {
				System.err.println("Receive: Sleep call failed.");
			} finally {
				
				State="Waiting ACK";
				
				endtime();

				if(lastTimeAckReceived<System.currentTimeMillis())
					lastTimeAckReceived=System.currentTimeMillis();
				
				showAck=true;
				
				/* sleep until no ack received anymore */
				while (showAck && (System.currentTimeMillis() - lastTimeAckReceived) < sleepAfterSend) {
					try {
						Thread.sleep(2000);
					} catch (InterruptedException e1) {
						System.out.println("" + e1.toString());
					}
				}
				closeSQFile();
				
				savestats(i);
				
				report("\n");
				
				try {
					if (MSock != null)
						MSock.close();
				} catch (Exception e) {
					report("Transmit: socket close error: " + e);
				}
				MSock = null;
				
				State="End";
			}
		}

	}

	I_OverlaySocket OLSock = null;

	/* The configuration object contains configuration  parameters */
	HyperCastConfig ConfObj = null;
	
	String overrides = "";

	void initOLSocket(I_ReceiveCallback cb) {
		/* Create the overlay manager using the configuration file */
		boolean filefound = false;
		
		if(Params.getOLConfig().equals("none")){
			report("Waiting to set ConfigFile...");
		}	
		while(Params.getOLConfig().equals("none")){
			
			try {
				Thread.sleep(3000);
			} catch (InterruptedException e1) {
				report(e1.toString());
			}
		}
		
		while (!filefound) {
			try {
				//TODO: no overrides???
				ConfObj = HyperCastConfig.createConfig(Params.getOLConfig());
				//OvManager = new OverlayManager(Params.getOLConfig(), overrides);
				filefound = true;

				/*
				 * Set security properties in the configuration onject.
				 */
				TransThread.setSecurityProperty(ConfObj);
				
			} catch (Exception e) {
				e.printStackTrace();
				if (!monitor)
					break;
				if (!Params.getTR() && !this.transStarted)
					break;

				try {
					Thread.sleep(5000);
				} catch (InterruptedException e1) {
					report(e1.toString());
				}

			}

		}
		
		if (OLSock == null) {

			OLSock = ConfObj.createOverlaySocket(cb, eh, this);
			//socketTable.add(OLSock);
			
			joined = true;
			OLSock.joinOverlay();

		}

		System.out.println(
			"This node's logical address is:" + OLSock.getLogicalAddress().toString());
		
		//TODO: not supported
		if(Params.FixKey){
			System.out.println("Fix the neighbor key");
			try{
				//OLSock.setStats("SP.FixKey", "true");
			}catch(Exception e){
				System.out.println(e.toString());
			}
		}

	}
	boolean joined = false;
	long lastTimeAckReceived = 0;
	boolean showAck=false;
	boolean readAck=true;
	

	public void ackReceived(String receiver, TTCPMessage TtcpMsg) {
		
		if(!readAck)return;
		
		lastTimeAckReceived = System.currentTimeMillis();
		saveDelay(
			receiver,
			TtcpMsg.getSeqNum(),
			TtcpMsg.ackSize,
			TtcpMsg.getTimestamp(),
			lastTimeAckReceived,
			TtcpMsg.receivedTime);
		
		
		if (showAck && numOfAck % 50 == 0) {
			System.out.print(">");
		}
		
		if(TtcpMsg.getSeqNum()>=(NumBuf-1)){
			report("\nLast ACK received.");
			showAck=false;
			readAck=false;
		}
		
	}

	//public void messageArrived(I_OverlayMessage RecvdMessage) {
	public void ReceiveCallback (I_OverlayMessage RecvdMessage) {
		/* callback */
		/* receive acks */
		TTCPMessage TtcpMsg = new TTCPMessage(RecvdMessage.getPayload());

		if (TtcpMsg.getType() != TTCPMessage.MSGTYPE_ACK)
			return;

		//System.out.println("ACK received: "+TtcpMsg.getSeqNum()+" "+TtcpMsg.getTimestamp()+"size="+TtcpMsg.ackSize);
		ackReceived(RecvdMessage.getSourceAddress().toString(), TtcpMsg);

	}
	I_LogicalAddress srcAddr=null;
	void responseOLMessage(I_OverlayMessage OlMsg, TTCPMessage TtcpMsg, long now) {
		//System.out.println("send response");

		TTCPMessage ResMsg = new TTCPMessage(TTCPMessage.MSGTYPE_ACK);
		ResMsg.setSeqNum(TtcpMsg.getSeqNum());
		ResMsg.setTimestamp(TtcpMsg.getTimestamp());
		ResMsg.ackSize = OlMsg.getPayloadLength();
		
		/*
		if (zeroTime == 0)
			zeroTime = now;
			*/
		
		ResMsg.receivedTime = now - zeroTime;

		byte[] payload = ResMsg.getPayload();
		I_OverlayMessage SendMessage = OLSock.createMessage(payload);

		/* send three times to make sure the receiver receives */
		if(srcAddr==null){
			srcAddr=OlMsg.getSourceAddress();
		}
		OLSock.sendToNode(SendMessage, srcAddr);

	}
	// Transfer using Overlay
	long totalDelay = 0;
	int numOfAck = 0;
	int numOfMsg = 0;

	private void TransOverlay() {
		int i = 0;
		long start_time = 0;
		long end_time = 0;
		long fix_sleep_time = 0;

		I_LogicalAddress LAs[] = null;
		
		if (TransmitReceive == ParamsClass.RECEIVE) {
			State="Wait";
			try {
				report("Overlay Receive: nbuf= " + NumBuf);

				initOLSocket(null);
				OLSock.setSoTimeout(2000);
				
				//Synchronization time with receiver
				report("Wait to receive multicast packets:");

				TTCPMessage TtcpMsg = new TTCPMessage((byte) 0);
				srcAddr=null;

				while (true) {
					I_OverlayMessage RecvdMessage = null;
					while (RecvdMessage == null && !toRestart)
						RecvdMessage = OLSock.receive();

					if (toRestart) {
						//System.out.println("\n******* Overlay socket restarted! ******* 1\n");
						return;
					}
					if (RecvdMessage == null)
						throw new InterruptedIOException();

					//System.out.println("The payload length is " + RecvdMessage.getPayloadLength ());
					if (RecvdMessage.getPayloadLength() > 1) {
						starttime();
						endtime();
						
						//int seq=getSequenceNumber(RecvdMessage.getPayload(), 0);
						//if(seq>=0 && seq<NumBuf)saveSQ(seq);

						TtcpMsg.restoreMessage(RecvdMessage.getPayload());
						
						String ss=TtcpMsg.getSession();
						if(ss!=null){
							session=ss;
							report("Session set to: "+ss);
						}
						
						
						
						if (TtcpMsg.getType() == TTCPMessage.MSGTYPE_DATA) {
							
							int seq = TtcpMsg.getSeqNum();
							
							if (seq >= 0 && seq < NumBuf)
								saveSQ(seq, TtcpMsg.getTimestamp(), 0, (OL_Message)RecvdMessage);
							if (TtcpMsg.getResponse())
								responseOLMessage(RecvdMessage, TtcpMsg, end);
							if (TtcpMsg.numOfPacket > 0) {
								NumBuf = TtcpMsg.numOfPacket;
								report("Overlay Receive: nbuf adjusted to " + NumBuf);
							}
							
							
							Timeout=(int) TtcpMsg.receivedTime;
							report("Timeout set to: "+Timeout);
							
							if(seq>=(NumBuf-1)){
								//maybe duplicate packet from the last session
								continue;
							}
							

						}
						break;
					} else {
						starttime();
						//OLSock.setSoTimeout(20000);
					}

				}

				stopReceiving = false;
				//starttime();
				if (autoStop)
					OLSock.setSoTimeout(Timeout);
					
				openSQFile(counter);
				openBWFile(counter);
				initBWCheck();

				report("Start to receive overlay packets:");
				eh.eventOccurred(new TTCP_EVENT(start, "Start receiving."));
				
				State="Receiving";
				for (i = 1; i < NumBuf; i++) {
					I_OverlayMessage RecvdMessage = OLSock.receive();

					if (RecvdMessage == null) {
						if (autoStop || stopReceiving)
							throw new InterruptedIOException();
						else {
							i--;
							continue;
						}
					}

					//int seq=getSequenceNumber(RecvdMessage.getPayload(), 0);
					//if(seq>=0 && seq<NumBuf)saveSQ(seq);

					endtime();
					checkBW();

					TtcpMsg.restoreMessage(RecvdMessage.getPayload());

					if (TtcpMsg.getType() != TTCPMessage.MSGTYPE_DATA)
						continue;

					int seq = TtcpMsg.getSeqNum();
					if (seq >= 0 && seq < NumBuf)
						saveSQ(seq, TtcpMsg.getTimestamp(), end, (OL_Message)RecvdMessage);
					
					if (TtcpMsg.getResponse()){
					
						responseOLMessage(RecvdMessage, TtcpMsg, end);
						
						if(seq>= (NumBuf-1)){
							/** send two more messages if this is the last msg */
							responseOLMessage(RecvdMessage, TtcpMsg, end);
							responseOLMessage(RecvdMessage, TtcpMsg, end);
						}
							
					}
					
					
					//if ((i%50 == 0) && (i > 0)) {
					if (i % 50 == 0) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
					
					if(seq>= (NumBuf-1) ){
						/* the last message */
						i++;
						break;
					}
					
				}
				System.out.println(".");
			} catch (InterruptedIOException e) {
				//System.err.println("Receive: Timeout when receiving data: " + e);
				System.err.println("Receive: Receiving finishes but not all packets received.");
			} catch (IOException e) {
				System.err.println("Receive: IO exception: " + e);
			} finally {

				closeBWFile();
				closeSQFile();

				//get retransmission packet number
				int BufSent = 0;
				try {
					//TODO
					//BufSent += Integer.parseInt(OLSock.getStats("Adapter.UPacketsSent"));
				} catch (Exception e) {
					//report(e.toString());	
				}
				try {
					//TODO
					//BufSent += Integer.parseInt(OLSock.getStats("Adapter.MPacketsSent"));
				} catch (Exception e) {
					//report(e.toString());	
				}
				//endtime();
				Params.BufSent = BufSent;
				savestats(i);
				try {
					if (OLSock != null && !monitor) {
						//joined = false;
						//OLSock.leaveGroup();
						//OLSock = null;
					} else if (monitor && toRestart) {
						System.out.println("\n******* Overlay socket restarted! ******* 2\n");

					}

				} catch (Exception e) {
					report("Receive: socket close error: " + e);
				}
				
				/* wait some time before finish */
				System.out.println("Waiting 5sec before finish...");
				try {
					Thread.sleep(5000);
				} catch (InterruptedException e1) {
					report(e1.toString());
				}
				State="End";

			}

		} else { // TRANSMIT Multicast

			if (rate == 0) {
				fix_sleep_time = 0;
			} else {
				fix_sleep_time = (new Float((BufSize * 8 * pkgs_onetime) / rate)).longValue();
				//System.out.println("fix_sleep_time=" + fix_sleep_time);
			}

			for (i = 0; i < BufSize; i++) {
				buf[i] = (byte) 123;
			}
			try {
				report("Overlay Transmit: nbuf= " + NumBuf);

				initOLSocket(this);

				String hostList[] = Params.getHostList();

				LAs = new I_LogicalAddress[hostList.length];

				for (int j = 0; j < LAs.length; j++) {
					report("Send to " + hostList[j]);
					LAs[j] = OLSock.createLogicalAddress(hostList[j]);
				}

				/* wait some time until the overlay converge */
				System.out.println("Waiting 5sec before overlay network becomes stable...");
				Thread.sleep(5000);

				//OLSock.setSoTimeout (2000);
			} catch (InterruptedException e) {
				System.err.println("Transmit: Interruption exception: " + e);
			}

			TTCPMessage TtcpMsg = new TTCPMessage(TTCPMessage.MSGTYPE_SYNC);
			byte[] payload = TtcpMsg.getPayload();
			I_OverlayMessage SendMessage = OLSock.createMessage(payload);

			/* send three times to make sure the receiver receives */
			//OLSock.sendToAll(SendMessage);
			//OLSock.sendToAll(SendMessage);
			//OLSock.sendToAll(SendMessage);

			TtcpMsg = new TTCPMessage(TTCPMessage.MSGTYPE_DATA, BufSize, 0);
			TtcpMsg.numOfPacket = NumBuf;
			TtcpMsg.setSession(session);
			TtcpMsg.receivedTime=Timeout;

			if (Params.getDelayFile() != null)
				TtcpMsg.setResponse(true);

			openDelayFile(counter);
			totalDelay = 0;
			numOfAck = 0;

			try {
				report("Start to send overlay packets");
				eh.eventOccurred(new TTCP_EVENT(System.currentTimeMillis(), "Start transmitting."));

				rate = Params.getRate();
				end_rate = Params.getEndRate();
				rate_step = Params.getRateStep();
				int rate_interval = 0;

				if (end_rate < rate) {
					rate_step = -rate_step;
				}
				if (rate > 0 && rate_step != 0)
					rate_interval =
						NumBuf
							* BufSize
							* 8
							* 2
							/ (end_rate + rate)
							/ ((end_rate - rate) / rate_step + 1);

				int start_index = 0;

				start_time = System.currentTimeMillis();
				starttime();
				
				openSQFile(counter);
				
				State="Transmitting";
				toStop=false;
				
				for (i = 0; i < NumBuf; i++) {
					//start_time = System.currentTimeMillis();
					/* sequence number */
					//setSequenceNumber(i, buf, 0);
					//SendMessage = OLSock.createMessage(buf, buf.length);
					if(toStop){
						throw new InterruptedException();
					}
					if (toRestart)
						throw new InterruptedException();

					TtcpMsg.setSeqNum(i);
					TtcpMsg.setTimestamp(System.currentTimeMillis());
					payload = TtcpMsg.getPayload();
					SendMessage = OLSock.createMessage(payload);
					if (LAs == null || LAs.length <= 0)
					{
						OLSock.sendToAll(SendMessage);
						//OLSock.sendToNode(SendMessage, OLSock.createLogicalAddress("60,60"));
					}
					else {
						//overlay unicast
						for (int j = 0; j < LAs.length; j++) {
							OLSock.sendToNode(SendMessage, LAs[j]);
							if(i>= (NumBuf-1)){
								/** send two mores if this is the last */
								OLSock.sendToNode(SendMessage, LAs[j]);
								OLSock.sendToNode(SendMessage, LAs[j]);
							}
						}
					}

					end_time = System.currentTimeMillis();
					saveSQ(i, end_time, end_time, (OL_Message)SendMessage);
					
					if (rate > 0) {
						long next_time = (i - start_index + 1) * BufSize * 8 / rate;
						long sleep_time = next_time - (end_time - start_time);
						//System.out.println("The " + i + "th packet is sent out!");
						//System.out.println("sleep time=" + sleep_time);
						if ((rate > 0) && (i % pkgs_onetime == 0) && ((i - start_index) > 0)) {
							if (sleep_time > 0) {
								//System.out.println("sleep time=" + (fix_sleep_time- end_time + start_time));
								Thread.sleep(sleep_time > 10 ? sleep_time - 10 : sleep_time);
							} else {
								//System.out.
								//println ("The specified sending rate is too fast!");
							}
							//System.out.println("sleep last time=" + (System.currentTimeMillis() - end_time));
							if (rate_interval > 0 && (end_time - start_time) >= rate_interval) {
								rate += rate_step;
								report("Rate steps to " + rate + " Kbps");
								start_index = i;
								start_time = System.currentTimeMillis();
							}
						}
					}

					if ((i % 50 == 0) && (i > 0)) {
						System.out.print(".");
						if(i % 1000 ==0)
							System.out.println("");
					}
				}
				System.out.println(".");
				System.out.println("finished.");

			} catch (InterruptedException e) {
				System.err.println("Receive: Sleep call failed.");
			} finally {
				
				State="Waiting ACK";
				
				endtime();
				
				if(lastTimeAckReceived<System.currentTimeMillis())
					lastTimeAckReceived=System.currentTimeMillis();
				
				showAck=true;
				
				if(!toStop){
					/* sleep until no ack received anymore */
					while (showAck && (System.currentTimeMillis() - lastTimeAckReceived) < sleepAfterSend) {
						try {
							Thread.sleep(2000);
						} catch (InterruptedException e1) {
							System.out.println("" + e1.toString());
						}
					}
				}
				

				//savestats (i);
				savestats(numOfAck);
				closeDelayFile();
				closeSQFile();

				try {
					/* wait some time before leaving */
					System.out.println("\nWaiting 5sec before leaving overlay network...");
					Thread.sleep(5000);
				} catch (InterruptedException e) {

				}
				try {
					if (OLSock != null && !monitor) {
						//joined = false;
						//OLSock.leaveGroup();
						//OLSock = null;
					} else if (monitor && toRestart) {
						System.out.println("\n******* Overlay socket restarted! ******* 3\n");

					}
				} catch (Exception e) {
					report("Transmit: socket close error: " + e);
				}
				
				State="End";
			}
		}

	}

	public static void setSecurityProperty(HyperCastConfig ConfObj)
	{

		/*
		 * The following variables are provided by application.
		 */
		String keystorepassword = "password";
		String privatekeyalias = "testpair";
		String privatekeypassword = "password";

		//HyperCastConfig ConfObj = HyperCastConfig.createOLConfig(args[0]);

		ConfObj.setPrivateTextAttribute(XmlUtil.createXPath("/Private/KeyStorePassword"), keystorepassword);
		ConfObj.setPrivateTextAttribute(XmlUtil.createXPath("/Private/PrivateKeyAlias"), privatekeyalias);
		ConfObj.setPrivateTextAttribute(XmlUtil.createXPath("/Private/PrivateKeyPassword"), privatekeypassword);


		/*
		 * If /Public/Security/SecurityLevel is set to GroupKeys, the group key needs to be set
		 * in below three steps.
		 */

		//Obtain the group key.
		SecretKeySpec groupKey;

		/*
		 * A byte array (obtained by other way) is used to generate the group key. In this example,
		 * the cypher algorithm specified in the configuration file by the attribute 
		 * /Public/Security/CryptAlgorithm is AES that needs 16-byte byte array to create the 
		 * secret key.
		 */
		byte[] aesData =
		 {(byte)0x7e,(byte)0x1a,(byte)0xab,(byte)0x55,(byte)0x4a,(byte)0x6f,(byte)0x11,(byte)0x79,(byte)0x2e,(byte)0x1e,
		  (byte)0x3b,(byte)0x55,(byte)0x3e,(byte)0xa1,(byte)0xc3,(byte)0x4b};

		groupKey = new SecretKeySpec(aesData, "AES");

		//Convert the group key into a string.
		String newInput;
		try
		{
			ByteArrayOutputStream bos = new ByteArrayOutputStream();
		
			ObjectOutputStream ketStringObj =
				new ObjectOutputStream(bos);
				
			ketStringObj.writeObject(groupKey);

			newInput = bos.toString();

			ketStringObj.close();
		} catch (Exception e) {
			throw new HyperCastFatalRuntimeException ("Error reading group key.");
		}

		//Set the group key into private attribute document..
		ConfObj.setPrivateTextAttribute(XmlUtil.createXPath("/Private/GroupKey"), newInput);

	}



	// Start and end a timed run and save stats.
	private void starttime() {
		//Date date = new Date ();
		//start = date.getTime ();
		start = System.currentTimeMillis();
	}
	private void endtime() {
		//Date date = new Date ();
		//end = date.getTime ();
		end = System.currentTimeMillis();
	}
	private void savestats(int i) {
		Params.setBufCount(i);
		Params.setStart(start);
		Params.setEnd(end);

		Params.setTotalDelay(totalDelay);

	}
	private void report(String s) {
		System.err.println(s);
	}
	String session = "TEST";
	boolean autoStop = true;
	boolean stopReceiving = false;
	
/*	
	private void initStatisticsStructure()
    {
        statsProcessor = new StatsProcessor();
		
        StatsElement se;

        se = new StatsElement ("NumOfSockets", null, this, true, false);
        se.setSchemaProperty ("xsd:Integer", 0, 1, null, null);
        statsProcessor.addStatsElement (se);

        se = new StatsElement ("CurrentTime", null, this, true, false);
        se.setSchemaProperty ("xsd:Long", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
       
        se = new StatsElement ("KillServer", null, this, false, true);
        se.setSchemaProperty ("xsd:Boolean", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("State", null, this, true, false);
        se.setSchemaProperty ("xsd:String", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("AutoStop", null, this, true, true);
        se.setSchemaProperty ("xsd:Boolean", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("Session", null, this, true, true);
        se.setSchemaProperty ("xsd:String", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("Timeout", null, this, false, true);
        se.setSchemaProperty ("xsd:Integer", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("SleepAfterSend", null, this, false, true);
        se.setSchemaProperty ("xsd:Integer", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("StopReceiving", null, this, false, true);
        se.setSchemaProperty ("xsd:Boolean", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("Stop", null, this, false, true);
        se.setSchemaProperty ("xsd:Boolean", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("Restart", null, this, false, true);
        se.setSchemaProperty ("xsd:Boolean", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("Transmit", null, this, false, true);
        se.setSchemaProperty ("xsd:Integer", 0, 1, null, null);
        statsProcessor.addStatsElement (se);

        if (socketTable == null)
            throw new RuntimeException ("socketTable must be non-null for statistics initialization");

        se = new StatsElement ("SocketTable", socketTable, null, true, true);
        se.setSchemaProperty ("xsd:complexType", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
        
        se = new StatsElement ("Params", Params, null, true, true);
        se.setSchemaProperty ("xsd:complexType", 0, 1, null, null);
        statsProcessor.addStatsElement (se);
          
    }
*/	
	
	/* (non-Javadoc)
	 * @see edu.virginia.cs.mng.hypercast.I_Stats#getStats(java.lang.String)
	 */
/*
	public String getStatsValue(String name) throws HyperCastStatsException {
		
		if (name.equals("State")) {
			return State;
		} else if (name.equals("AutoStop")) {
			return "" + autoStop;
		} else if (name.equals("Session")) {
			return session;
		} else if (name.equals("NumOfSockets")) {
			return "" + socketTable.size();
		} else if (name.equals("CurrentTime"))
			return "" + System.currentTimeMillis();
		else
			throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);
	}
*/

	/* (non-Javadoc)
	 * @see edu.virginia.cs.mng.hypercast.I_Stats#setStats(java.lang.String, java.lang.String)
	 */
	boolean toRestart = false;
	boolean toStop = false;
	long sleepAfterSend = 10000;

/*
	public String setStatsValue(String name, String value) throws HyperCastStatsException {
		
		if (name.equals("Timeout")) {

			Timeout = Integer.parseInt(value);

		} else if (name.equals("SleepAfterSend")) {

			sleepAfterSend = Long.parseLong(value);

		} else if (name.equals("Session")) {
			session = value;
		} else if (name.equals("AutoStop")) {
			if (value.equals("true"))
				autoStop = true;
			else
				autoStop = false;
		} else if (name.equals("StopReceiving")) {
			if (value.equals("true"))
				stopReceiving = true;
			else
				stopReceiving = false;
		}else if (name.equals("Stop")) {
			if (value.equals("true")) {

					toStop = true;

				} else
					toStop = false;
		}  
		else if (name.equals("Restart")) {
			if (Params.getProtocol() == ParamsClass.OVERLAY) {
				if (value.equals("true")) {

					if (!transStarted && !Params.getTR()) {
						System.out.println("\n******* Overlay socket restarted! ******* 4\n");

						if (OLSock != null)
							OLSock.leaveOverlay();
						OLSock = null;
						initOLSocket(this);

					} else
						toRestart = true;

				} else
					toRestart = false;
			}
		} else if (name.equals("Transmit")) {

			//System.out.println("Start transmit");

			if (Params.getTR())
				throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);
			
			int rounds = Integer.parseInt(value);

			//System.out.println("start transmit socket-"+socketIndex+" "+rounds+" times");

			Params.setStop(false);
			//for(int i=0;i<rounds;i++)
			Params.setRepeatTime(rounds);
			(new Thread(this)).start();

		} else if (name.equals("KillServer")) {
			TimeBomb b = new TimeBomb(2000);

		} else
			throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);
		
		return value;
	}
*/	

	class EventHandler extends NotificationHandler{
		/* (non-Javadoc)
		 * @see edu.virginia.cs.mng.hypercast.Notification_Handler#handler(java.lang.Object)
		 */
		//public void handler(Object obj) {
		public void handler(NOTIFICATION_EVENT obj) {
			//System.out.println("Event here!!"+obj);
			saveEvent(((NOTIFICATION_EVENT)obj).getTimestamp(), (NOTIFICATION_EVENT)obj);
/*			
			if (obj instanceof TTCP_EVENT)
			{
				saveEvent(((TTCP_EVENT)obj).getTimestamp(), (TTCP_EVENT)obj);
			}
*/			
		}
	}
	
	/* (non-Javadoc)
	 * @see edu.virginia.cs.mng.hypercast.I_InterceptionCallBack#messageArrived(edu.virginia.cs.mng.hypercast.OL_Message)
	 */

	//public OL_Message messageArrived(OL_Message msg) {
	public I_OverlayMessage InterceptionCallback(I_OverlayMessage msg) {
		//System.out.println("Intercept");
		this.saveIntercept(System.currentTimeMillis(), (OL_Message) msg);
		return msg;
	}

	public OL_Message messageArrived(OL_Message msg) {
		//System.out.println("Intercept");
		this.saveIntercept(System.currentTimeMillis(), msg);
		return msg;
	}


	 /** @see I_Stats#getStats */
/*
    public Element getStats (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getStatsResult (doc, xpath);
    }
*/
    /** @see I_Stats#setStats */
/*
    public Element setStats (final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException
    {
        return statsProcessor.setStatsResult (doc, xpath, newValue);
    }
*/

    /** @see I_Stats#getReadSchema */
/*
    public Element getReadSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getReadSchemaResult (doc, xpath);
    }
*/

    /** @see I_Stats#getWriteSchema */
/*
    public Element getWriteSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getWriteSchemaResult (doc, xpath);
    }
*/	

}
class TimeBomb implements Runnable {
	int millisecondsToExplosion;

	TimeBomb(int t) {
		millisecondsToExplosion = t;
		Thread th = new Thread(this);
		th.setPriority((Thread.MAX_PRIORITY + Thread.NORM_PRIORITY) / 2);
		th.start();
	}

	public void run() {
		try {
			Thread.sleep(millisecondsToExplosion);
		} catch (Exception e) {
		}

		System.exit(0);
	}
}

